package checkers.igj;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;
import javax.lang.model.type.*;

import com.sun.source.tree.*;
import com.sun.source.util.SimpleTreeVisitor;
import com.sun.source.util.TreePath;

import checkers.igj.quals.I;
import checkers.igj.quals.Immutable;
import checkers.igj.quals.Mutable;
import checkers.igj.quals.ReadOnly;
import checkers.types.*;
import checkers.types.AnnotatedTypeMirror.*;
import checkers.util.AnnotationUtils;
import checkers.util.TreeUtils;
import checkers.util.TypesUtils;

/**
 * Generates {@link AnnotatedTypeMirror} instances accounting for the properties
 * of the IGJ type system. For example, when the
 * {@link IGJAnnotatedTypeFactory#getAnnotatedType(Tree)} method is invoked
 * on a primitive int, it returns the type "@Immutable int".
 */
public class IGJAnnotatedTypeFactory extends AnnotatedTypeFactory{
    /** Adds annotations from tree context before type resolution. */
    private final TreePreAnnotator treePre;

    /** Adds annotations from the resulting type after type resolution. */
    private final TypePostAnnotator typePost;

    /** The various IGJ annotation. */
    protected final AnnotationMirror READONLY, MUTABLE, IMMUTABLE, I,
    PLACE_HOLDER, ASSIGNABLE, ASSIGNS_FIELDS;

    public IGJAnnotatedTypeFactory(ProcessingEnvironment env,
            CompilationUnitTree root) {
        super(env, root);
        treePre = new TreePreAnnotator(env, root);
        typePost = new TypePostAnnotator(env, root);
        AnnotationFactory annoFactory = new AnnotationFactory(env);
        READONLY = annoFactory.fromName(ReadOnly.class.getCanonicalName());
        MUTABLE = annoFactory.fromName(Mutable.class.getCanonicalName());
        IMMUTABLE = annoFactory.fromName(Immutable.class.getCanonicalName());
        I = annoFactory.fromName(I.class.getCanonicalName());
        PLACE_HOLDER =
            annoFactory.fromName(IGJPlaceHolder.class.getCanonicalName());
        ASSIGNABLE = annoFactory.fromName("checkers.igj.quals.Assignable");
        ASSIGNS_FIELDS =
            annoFactory.fromName("checkers.igj.quals.AssignsFields");

    }

    /**
     * Returns the annotation specifying the immutability type of {@code type}.
     */
    public AnnotationMirror getIGJAnnotation(AnnotatedTypeMirror type) {
        if (type.hasAnnotation(I))
            return type.getAnnotation(AnnotationUtils.annotationName(I));
        if (type.hasAnnotation(READONLY))
            return READONLY;
        else if (type.hasAnnotation(MUTABLE))
            return MUTABLE;
        else if (type.hasAnnotation(IMMUTABLE))
            return IMMUTABLE;
        else if (type.hasAnnotation(PLACE_HOLDER))
            return PLACE_HOLDER;
        else if (type.hasAnnotation(ASSIGNS_FIELDS))
            return ASSIGNS_FIELDS;
        else
            return null;
    }
    
    @Override
    public AnnotatedTypeMirror getAnnotatedType(final Tree tree) {
        
        // Get the underlying expression if the input is parenthesized.
        Tree t = skipParens(tree);
        
        // Get annotations based on tree properties.
        List<AnnotationMirror> preAnnotations = treePre.visit(t,
            new LinkedList<AnnotationMirror>());
        
        AnnotatedTypeMirror result = super.getAnnotatedType(t);
        result.addAnnotations(preAnnotations);
        immtabilityHackFixer.visit(result);
        
        // Try to resolve immutability type variable
        if (t instanceof ExpressionTree)
            resolveImmutabilityTypeVar(result, (ExpressionTree) t);
        
        // Add annotations based on type properties.
        typePost.visit(result);
        
        return result;
    }
    
    public boolean resolveImmutabilityTypeVar(AnnotatedTypeMirror type,
            ExpressionTree tree) {
        // TODO: Clear up this
        AnnotatedTypeMirror receiver = getReceiver(tree);
        ImmutabilityTemplateCollector collector =
            new ImmutabilityTemplateCollector();
        final Set<AnnotatedTypeMirror> types = 
            new HashSet<AnnotatedTypeMirror>();
        
        Map<String, AnnotationMirror> templateMapping = 
            collector.visit(receiver);

        // There is nothing to resolve
        if (templateMapping == null || templateMapping.isEmpty()) 
            return false;
        
        new ImmutabilityResolver().visit(type, templateMapping);
        
        if (tree instanceof MethodInvocationTree) {
            MethodInvocationTree methodInvoc = (MethodInvocationTree) tree;
            
            Map<String, AnnotationMirror> map = null;
            for (ExpressionTree param : methodInvoc.getArguments()) {
                AnnotatedTypeMirror tm = getAnnotatedType(param);
                map = collector.reduce(map, collector.visit(tm));
            }
            if (map != null && !map.isEmpty())
                new ImmutabilityResolver().visit(type, map);
        }
        
        // This is buggy
        new AnnotatedTypeScanner<Void, Void>() {
            @Override
            public Void visitDeclared(AnnotatedDeclaredType type, Void p) {
                if (type.hasAnnotation(I)) {
                    AnnotationMirror anno = 
                        type.getAnnotation(I.class.getCanonicalName());
                    AnnotationUtils annoUtil = new AnnotationUtils(env);
                    String s = annoUtil.parseStringValue(anno, 
                            IGJChecker.IMMUTABILITY_KEY);
                    if (!types.contains(s)) {
                        type.removeAnnotation(I);
                        type.addAnnotation(PLACE_HOLDER);
                    }
                }
                return super.visitDeclared(type, p);
            }
        }.visit(type);
        
        return true;
    }

    @Override
    public AnnotatedTypeMirror getAnnotatedType(Element elt) {
        
        AnnotatedTypeMirror result = super.getAnnotatedType(elt);
        immtabilityHackFixer.visit(result);
        // Add annotations based on type properties.
        typePost.visit(result);
        return result;
    }

    public AnnotatedExecutableType getMethodType(MethodInvocationTree methodInvocation) {
    	ExecutableElement elem = this.elementFromUse(methodInvocation);
    	
    	AnnotatedExecutableType method = 
    		(AnnotatedExecutableType) getAnnotatedType(elem);
    	
    	return method;
    }

    private class TreePreAnnotator
    extends SimpleTreeVisitor<List<AnnotationMirror>, List<AnnotationMirror>> {
    	
    	private final ProcessingEnvironment env;
    	private final CompilationUnitTree root;
    	private final AnnotatedTypeFactory factory;
    	
    	public TreePreAnnotator(ProcessingEnvironment env, 
    			CompilationUnitTree root) {
    		this.env = env;
    		this.root = root;
    		this.factory = new AnnotatedTypeFactory(this.env, this.root);
    	}
    	
    	public List<AnnotationMirror> visit(Tree node) {
    		return visit(node, new ArrayList<AnnotationMirror>());
    	}
    	
        @Override
        public List<AnnotationMirror> defaultAction(Tree node,
                List<AnnotationMirror> p) {
            return p;
        }

        @Override
        public List<AnnotationMirror> visitNewClass(NewClassTree node,
        		List<AnnotationMirror> p) {
        	// If no immutability is specified, use a place holder
        	AnnotatedDeclaredType dt = (AnnotatedDeclaredType)
        		factory.getAnnotatedType(node);
        	
        	if (!hasImmutabilityType(dt))
        		p.add(IGJAnnotatedTypeFactory.this.PLACE_HOLDER);
        	
        	return p;
        }
        
        @Override
        public List<AnnotationMirror> visitLiteral(LiteralTree node,
        		List<AnnotationMirror> p) {
        	// Literals are immutable by default
        	if (p.contains(IMMUTABLE))
        		p.add(IMMUTABLE);
        	return p;
        }
        
    }

    /**
     * A class for adding annotations to a type after initial type resolution.
     * Used here for resolving this-mutability.
     */
    private class TypePostAnnotator  extends AnnotatedTypeScanner<Void, Void> {

    	
    	private final ProcessingEnvironment env;
    	private final CompilationUnitTree root;
    	private final AnnotatedTypeFactory simpleFactory;
    	
    	public TypePostAnnotator(ProcessingEnvironment env, 
    			CompilationUnitTree root) {
    		this.env = env;
    		this.root = root;
    		this.simpleFactory = new AnnotatedTypeFactory(this.env, this.root);
    	}
    	
        /** All primitives are immutable **/
        @Override
        public Void visitPrimitive(AnnotatedPrimitiveType type, Void p) {
        	if (!hasImmutabilityType(type))
        		type.addAnnotation(IMMUTABLE);
            return super.visitPrimitive(type, p);
        }

        /** Returns {@code true} if type is a boxed primitive type **/
        protected boolean isBoxedImmutable(AnnotatedDeclaredType type) {
            DeclaredType underlyingType = type.getUnderlyingType();
            Name qualifiedName = TypesUtils.getQualifiedName(underlyingType);

            return (qualifiedName.contentEquals("java.lang.String") ||
                    qualifiedName.contentEquals("java.lang.Boolean") ||
                    qualifiedName.contentEquals("java.lang.Byte") ||
                    qualifiedName.contentEquals("java.lang.Character") ||
                    qualifiedName.contentEquals("java.lang.Short") ||
                    qualifiedName.contentEquals("java.lang.Integer") ||
                    qualifiedName.contentEquals("java.lang.Number") ||
                    qualifiedName.contentEquals("java.lang.Long") ||
                    qualifiedName.contentEquals("java.lang.Double") ||
                    qualifiedName.contentEquals("java.lang.Float") ||
                    qualifiedName.contentEquals("java.lang.Class") ||
                    qualifiedName.contentEquals("java.lang.Void"));
        }

        @Override
        public Void visitDeclared(AnnotatedDeclaredType type, Void p) {
        	if (!hasImmutabilityType(type)) {
        		// By default:
        		//  - Classes: 				Mutable
        		//  - Interfaces: 			Place holder
        		//  - Enum & Annotations:	Immutable
        		
        		Element elem = type.getUnderlyingType().asElement();
        		AnnotatedDeclaredType elemType = (AnnotatedDeclaredType)
        			simpleFactory.getAnnotatedType(elem);

        		if (isBoxedImmutable(type))
        			type.addAnnotation(IMMUTABLE);
        		else if (hasImmutabilityType(elemType)) {
        		    if (elemType.hasAnnotation(IMMUTABLE))
        		        type.addAnnotation(IMMUTABLE);
        		    else
        		        type.addAnnotation(MUTABLE);
        		} else if (elem.getKind() == ElementKind.CLASS)
        			type.addAnnotation(MUTABLE);
        		else if (elem.getKind() == ElementKind.INTERFACE)
        			type.addAnnotation(PLACE_HOLDER);
        		else if (elem.getKind() == ElementKind.ENUM ||
        				elem.getKind() == ElementKind.ANNOTATION_TYPE)
        			type.addAnnotation(IMMUTABLE);
        		else
        			assert false : "Shouldn't be here";
        	}
        	
        	return super.visitDeclared(type, p);
        }
        
        @Override
        public Void visitNull(AnnotatedNullType type, Void p) {
        	if (!hasImmutabilityType(type))
        		type.addAnnotation(PLACE_HOLDER);
        	
        	return super.visitNull(type, p);
        }
        
        @Override
        public Void visitExecutable(AnnotatedExecutableType type, Void p) {
        	ExecutableElement elem = (ExecutableElement) type.getElement();
        	
        	// Check Receiver
        	if (!hasImmutabilityType(type.getReceiverType())) {
        		AnnotatedDeclaredType receiverType = type.getReceiverType();
        		TypeElement classImmutability = 
        			(TypeElement) elem.getEnclosingElement();
        		AnnotatedDeclaredType classType = 
        			(AnnotatedDeclaredType) getAnnotatedType(classImmutability);
        		
        		if (elem.getKind() == ElementKind.CONSTRUCTOR)
        			receiverType.addAnnotation(ASSIGNS_FIELDS);
        		else if (classType.hasAnnotation(IMMUTABLE))
        			receiverType.addAnnotation(READONLY);
        		else if (classType.hasAnnotation(MUTABLE))
        			receiverType.addAnnotation(MUTABLE);
        		else 
        			receiverType.addAnnotation(MUTABLE);
        	}
        	
        	return super.visitExecutable(type, p);
        }
    }

    /**
     * @param type  an annotated type mirror
     * @return  true iff the type is specified an immutability type, 
     *          false otherwise
     */
    public boolean hasImmutabilityType(AnnotatedTypeMirror type) {
        return getIGJAnnotation(type) != null;
    }
    
    /**
     * A Helper class that tries to resolve the immutability type variable,
     * as the type variable is assigned to the most restricted immutability
     */
    private class ImmutabilityTemplateCollector
    extends AnnotatedTypeScanner<Map<String, AnnotationMirror>, Void> {
        
        @Override
        public Map<String, AnnotationMirror> reduce(Map<String, AnnotationMirror> r1,
                Map<String, AnnotationMirror> r2) {
            Map<String, AnnotationMirror> result = 
                new HashMap<String, AnnotationMirror>();
            
            if (r1 != null)
                result.putAll(r1);
            
            if (r2 != null) {
                // Need to be careful about overlap
                for (String key : r2.keySet()) {
                    if (!result.containsKey(key))
                        result.put(key, r2.get(key));
                    else if (!result.get(key).equals(r2.get(key)))
                        result.put(key, READONLY);
                }
            }
            return result;
        }
        
        @Override
        public Map<String, AnnotationMirror> visitDeclared(
                AnnotatedDeclaredType type, Void p) {
            Map<String, AnnotationMirror> result =
                new HashMap<String, AnnotationMirror>();
            
            TypeElement typeElement = 
                (TypeElement) type.getUnderlyingType().asElement();
            AnnotatedDeclaredType dcType = 
                (AnnotatedDeclaredType) getAnnotatedType(typeElement);
            
            if (dcType.hasAnnotation(I)) {
                AnnotationUtils annoUtil = new AnnotationUtils(env);
                String immutableString = 
                    annoUtil.parseStringValue(getIGJAnnotation(dcType), 
                            IGJChecker.IMMUTABILITY_KEY);
                AnnotationMirror immutability = getIGJAnnotation(type);
                // Assertion failes some times
//                assert immutability != null;
                if ((immutability != null) && (!type.hasAnnotation(ASSIGNS_FIELDS)))
                    result.put(immutableString, immutability);
            }

            return reduce(result, super.visitDeclared(type, p));
        }
    }
    
    /**
     * A helper class that resolves the immutability on a types based on a
     * provided mapping.
     * 
     * It returns a set of the annotations that were inserted. This is important
     * to recognize which immutability type variables were resolved and which
     * are to be made into place holder.
     */
    private class ImmutabilityResolver extends
    AnnotatedTypeScanner<Set<AnnotationMirror>, Map<String, AnnotationMirror>> {
        
        @Override
        public Set<AnnotationMirror> reduce(Set<AnnotationMirror> r1,
                Set<AnnotationMirror> r2) {
            Set<AnnotationMirror> result = new HashSet<AnnotationMirror>();
            if (r1 != null)
                result.addAll(r1);
            if (r2 != null)
                result.addAll(r2);
            return result;
        }
        
        @Override
        public Set<AnnotationMirror> visitDeclared(AnnotatedDeclaredType type, 
                Map<String, AnnotationMirror> p) {
            Set<AnnotationMirror> result = new HashSet<AnnotationMirror>();

            if (type.hasAnnotation(I)) {
                AnnotationUtils annoUtil = new AnnotationUtils(env);
                String immutableString = 
                    annoUtil.parseStringValue(getIGJAnnotation(type), 
                            IGJChecker.IMMUTABILITY_KEY);
                type.removeAnnotation(I);
                if (p.containsKey(immutableString)) {
                    type.addAnnotation(p.get(immutableString));
                    result.add(p.get(immutableString));
                }
            }
            
            return reduce(result, super.visitDeclared(type, p));
        }
    }

    ////////////// Copied Code
    protected AnnotatedDeclaredType getReceiver(ExpressionTree expression) {
        
        if (!(expression instanceof MethodInvocationTree
                || expression instanceof MemberSelectTree
                || expression instanceof IdentifierTree))
            // No receiver type for those
            return null;
        
        Tree receiver = expression;
        
        if (receiver instanceof MethodInvocationTree)
            receiver = ((MethodInvocationTree)receiver).getMethodSelect();

        // Figure out if this a self type
        if (receiver instanceof IdentifierTree)
            return getSelfType(expression);

        if (receiver instanceof MemberSelectTree)
            receiver = ((MemberSelectTree)receiver).getExpression();

        if (receiver instanceof IdentifierTree) {
            IdentifierTree iden = (IdentifierTree)receiver;
            if (iden.getName().contentEquals("this")
                    || iden.getName().contentEquals("super"))
                return getSelfType(expression);
        }
        return (AnnotatedDeclaredType) getAnnotatedType(receiver);
    }
    
    /**
     * Returns the type of {@code this} in the current location, which can
     * be used if {@code this} has a special semantics (e.g. {@code this}
     * is non-null)
     */
//    @Override
//    public AnnotatedDeclaredType getSelfType(Tree tree) {
//        // For now simply return the type of the enclosing class
//        ClassTree ct = TreeUtils.enclosingClass(TreePath.getPath(root, tree));
//        AnnotatedDeclaredType act = (AnnotatedDeclaredType) getAnnotatedType(ct);
//
//        MethodTree mt = TreeUtils.enclosingMethod(TreePath.getPath(root, tree));
//        if (mt != null) {
//            AnnotatedExecutableType amt = (AnnotatedExecutableType) getAnnotatedType(mt);
//            AnnotatedDeclaredType receiverType = amt.getReceiverType();
//            if (receiverType.hasAnnotation(READONLY) && act.hasAnnotation(I))
//                    return act;
//                else
//                    return receiverType;
//        }
//        return act;
//    }

    /**
     * Returns a type of the class and the field {@code this}.
     */
    @Override
    public AnnotatedDeclaredType getSelfType(Tree tree) {
        ClassTree ct = TreeUtils.enclosingClass(TreePath.getPath(root, tree));
        AnnotatedDeclaredType act = (AnnotatedDeclaredType) getAnnotatedType(ct);

        MethodTree mt = TreeUtils.enclosingMethod(TreePath.getPath(root, tree));
        if (mt == null)
            return act;

        AnnotatedExecutableType method = 
            (AnnotatedExecutableType) getAnnotatedType(mt);
        
        AnnotatedDeclaredType methodReceiver = method.getReceiverType();
        
        // Are we in a mutable or Immutable scope
        if (methodReceiver.hasAnnotation(MUTABLE) ||
                methodReceiver.hasAnnotation(IMMUTABLE)) {
            return methodReceiver;
        } else if (act.hasAnnotation(I)) {
            if (methodReceiver.hasAnnotation(ASSIGNS_FIELDS))
                act.addAnnotation(ASSIGNS_FIELDS);
            return act;
        } else if (act.hasAnnotation(IMMUTABLE)) {
            return act;
        } else // TODO: Figure out when
            return methodReceiver;
    }

    private AnnotatedTypeScanner<Void, Void> immtabilityHackFixer = 
        new AnnotatedTypeScanner<Void, Void>() {

        @Override
        public Void visitDeclared(AnnotatedDeclaredType type, Void p) {
            // Check if it has multiple annotations
            if (type.hasAnnotation(PLACE_HOLDER)) {
                type.removeAnnotation(READONLY);
                type.removeAnnotation(I);
                type.removeAnnotation(MUTABLE);
                type.removeAnnotation(IMMUTABLE);                
            } else if (type.hasAnnotation(READONLY)) {
                type.removeAnnotation(I);
                type.removeAnnotation(MUTABLE);
                type.removeAnnotation(IMMUTABLE);
            } else if (type.hasAnnotation(MUTABLE) || type.hasAnnotation(IMMUTABLE))
                type.removeAnnotation(I);
            return super.visitDeclared(type, p);
        }
    };
}
