package checkers.igj;

import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.Modifier;
import javax.lang.model.element.Name;

import com.sun.source.tree.*;

import checkers.basetype.BaseTypeVisitor;
import checkers.igj.quals.*;
import checkers.source.Result;
import checkers.types.AnnotatedTypeMirror;
import checkers.types.AnnotationFactory;
import checkers.types.AnnotatedTypeMirror.*;
import checkers.util.*;

/**
 * A type-checking visitor for the IGJ type qualifiers (
 * {@code @ReadOnly}, {@code @Immutable}, {@code @Mutable},
 * {@code @AssignsFields}, and {@code @Assignable}) that uses
 * the {@link BaseTypeVisitor} implementation.
 *
 * @see BaseTypeVisitor
 */
public class IGJVisitor extends BaseTypeVisitor<Void, Void> {

    protected final IGJAnnotatedTypeFactory factory;
    protected final AnnotationFactory annoFactory;
    protected final AnnotationMirror READONLY, MUTABLE, IMMUTABLE, I, 
        PLACE_HOLDER, ASSIGNABLE, ASSIGNS_FIELDS;

    public IGJVisitor(IGJChecker checker, CompilationUnitTree root) {
        super(checker, root);
        factory = new IGJAnnotatedTypeFactory(
                checker.getProcessingEnvironment(), root);
        annoFactory = new AnnotationFactory(checker.getProcessingEnvironment());
        READONLY = annoFactory.fromName(ReadOnly.class.getCanonicalName());
        MUTABLE = annoFactory.fromName(Mutable.class.getCanonicalName());
        IMMUTABLE = annoFactory.fromName(Immutable.class.getCanonicalName());
        I = annoFactory.fromName(I.class.getCanonicalName());
        PLACE_HOLDER = 
            annoFactory.fromName(IGJPlaceHolder.class.getCanonicalName());
        ASSIGNABLE = annoFactory.fromName("checkers.igj.quals.Assignable");
        ASSIGNS_FIELDS = 
            annoFactory.fromName("checkers.igj.quals.AssignsFields");
    }

    /**
     * Checks that a method invocation is a valid invocation, given the 
     * immutability scope it is in.
     */
    @Override
    public Void visitMethodInvocation(MethodInvocationTree node, Void p) {
    	
    	if (node.getArguments().isEmpty() 
    			&& node.getMethodSelect() instanceof IdentifierTree
    			&& ((IdentifierTree)node.getMethodSelect()).getName().contentEquals("super"))
    		return super.visitMethodInvocation(node, p);
    	
    	ExecutableElement elt = factory.elementFromUse(node);
    	if (elt == null || elt.getModifiers().contains(Modifier.STATIC))
    	    return super.visitMethodInvocation(node, p);
    	
        AnnotatedExecutableType method = factory.getMethodType(node);
        
        AnnotatedDeclaredType receiver = factory.getReceiver(node);
        
        if (!this.isInvokable(method, receiver, isSelfAccess(node))) {
            checker.report(Result.failure("methodInvocation.invalid",
                    method.getReceiverType().toString(),
                    receiver.toString()), node);
        }
        
        return super.visitMethodInvocation(node, p);
    }

    /**
     * Returns {@code true} if the tree represents an access to a member of
     * the same accessing instance.
     * 
     * Performs a compile-time check and would not handle run-time references
     * to the receiver instance. Thus, in the case:
     * 
     * <code>
     *      Object selfReference = this;
     *      int hashCode = selfReference.hashCode()
     * </code>
     * 
     * the hashCode() invocation wouldn't be considered a self access.
     * 
     * @param tree  expression tree representing an access to object member
     * @return {@code true} iff the member is a member of {@code this} instance
     */
    // I don't think that this method handles self references within private
    // inner classes, but IGJ Specification is unclear about it.
    protected boolean isSelfAccess(ExpressionTree tree) {
        ExpressionTree tr = tree;
        // If method invocation check the method select
        if (tree.getKind() == Tree.Kind.METHOD_INVOCATION) {
            tr = ((MethodInvocationTree)tree).getMethodSelect();
        }
        
        if (tr.getKind() == Tree.Kind.IDENTIFIER)
            return true;
        
        if (tr.getKind() == Tree.Kind.MEMBER_SELECT) {
            tr = ((MemberSelectTree)tr).getExpression();
            if (tr.getKind() == Tree.Kind.IDENTIFIER) {
                Name ident = ((IdentifierTree)tr).getName();
                return ident.contentEquals("this") ||
                        ident.contentEquals("super");
            }
        }

        return false;
    }
    
    /**
     * Checks the validity of the reassignability of a field
     */
    @Override
    protected void commonAssignmentCheck(Tree varTree, ExpressionTree valueExp, 
            String errorKey, Void p) {
        
        Element varElement = InternalUtils.symbol(varTree);
        if ((varElement.getKind() == ElementKind.FIELD) && 
                (varTree.getKind() != Tree.Kind.VARIABLE)) {
            // Check assignability
            ExpressionTree expTree = (ExpressionTree)varTree;
            AnnotatedTypeMirror variable = factory.getAnnotatedType(expTree);
            AnnotatedTypeMirror receiver = this.getReceiver(expTree);
            boolean isSelf = isSelfAccess(expTree);
            if (!isAssignable(variable, receiver, isSelf)) {
                checker.report(Result.failure("assignability.invalid"), varTree);
            }
        }
        
        // Now check assignability
        super.commonAssignmentCheck(varTree, valueExp, errorKey, p);
    }
    
    @Override
    protected void commonAssignmentCheck(AnnotatedTypeMirror varType,
            AnnotatedTypeMirror valueType, ExpressionTree valueTree,
            String errorKey, Void p) {
        super.commonAssignmentCheck(varType, valueType, valueTree, 
                "type.incompatible", p);
    }
    
    /**
     * Verifies that a field is assignable or not within a given scope.
     * 
     * IGJ Rule:
     * o.someField = ... is valid iff I(o) = Mutable or (I(o) =
     * AssignsFields and o = {@code this})
     * 
     * Exceptions: 1. Static Fields may be reassigned. 2. Assignable
     * fields may be reassigned all the time
     * 
     */
    protected boolean isAssignable(AnnotatedTypeMirror field, 
            AnnotatedTypeMirror receiver, boolean isSelfReference) {
        if (field.hasAnnotation(ASSIGNABLE) || 
                receiver.hasAnnotation(MUTABLE))
            return true;
        else if (isSelfReference && receiver.hasAnnotation(ASSIGNS_FIELDS))
            return true;
        else
            return false;
    }
    
    /**
     * Checks if the method in invokable with respect to IGJ Method Invocation
     * Rule.
     */
    protected boolean isInvokable(AnnotatedExecutableType method,
            AnnotatedTypeMirror receiver, boolean isSelfReference) {
        AnnotatedTypeMirror methodReceiver = method.getReceiverType();
        if (methodReceiver.hasAnnotation(READONLY)) {
            return true;
        } else if (methodReceiver.hasAnnotation(MUTABLE)) {
            return receiver.hasAnnotation(MUTABLE);
        } else if (methodReceiver.hasAnnotation(IMMUTABLE)) {
            return receiver.hasAnnotation(IMMUTABLE);
        } else if (methodReceiver.hasAnnotation(ASSIGNS_FIELDS)) {
            return (isSelfReference && receiver.hasAnnotation(ASSIGNS_FIELDS))
                || receiver.hasAnnotation(MUTABLE);
        }
        
        assert false : "We shouldn't be here!";
        return false;
    }
}
